use diesel::prelude::*;
use rustracing_jaeger::Span;

use crate::{db::manager::DB_MANAGER, metrics};

pub struct DBConnection {
    db: super::RawDBConnection
}

impl From<super::RawDBConnection> for DBConnection {
    fn from(conn: super::RawDBConnection) -> Self { Self { db: conn } }
}

impl DBConnection {
    // Users
    pub fn create_user_password(&mut self, span: &Span, name: String, password: String) -> super::User {
        let _span = metrics::span("create_user_password_in_db", span);
        let mut new_user: super::User = diesel::insert_into(crate::schema::user::table)
            .values(super::user::NewUser {
                name,
                password,
                gitlab: false,
                role: super::UserRole::Disabled,
                root_id: 0,
                tfa_type: super::TfaTypes::None,
                tfa_secret: None,
                gitlab_at: None,
                gitlab_rt: None
            })
            .get_result(&mut self.db)
            .expect("Failed to insert new user");

        let root_node = crate::routes::fs::create_node(span, "".to_owned(), &new_user, false, None, true, self)
            .expect("Couldn't create root node");
        new_user.root_id = root_node.id;
        self.save_user(span, &new_user);
        new_user
    }

    pub fn create_user_gitlab(
        &mut self,
        span: &Span,
        name: String,
        role: super::UserRole,
        gitlab_at: String,
        gitlab_rt: String
    ) -> super::User {
        let _span = metrics::span("create_user_gitlab_in_db", span);
        let mut new_user: super::User = diesel::insert_into(crate::schema::user::table)
            .values(super::user::NewUser {
                name,
                password: "".to_owned(),
                gitlab: true,
                role,
                root_id: 0,
                tfa_type: super::TfaTypes::None,
                tfa_secret: None,
                gitlab_at: Some(gitlab_at),
                gitlab_rt: Some(gitlab_rt)
            })
            .get_result(&mut self.db)
            .expect("Failed to insert new user");

        let root_node = crate::routes::fs::create_node(span, "".to_owned(), &new_user, false, None, true, self)
            .expect("Couldn't create root node");
        new_user.root_id = root_node.id;
        self.save_user(span, &new_user);
        new_user
    }

    pub fn get_user(&mut self, span: &Span, _id: i32) -> Option<super::User> {
        use crate::schema::user::dsl::*;
        let _span = metrics::span("get_user_from_db", span);
        user.find(_id).first(&mut self.db).ok()
    }

    pub fn find_user(&mut self, span: &Span, _name: &str, _gitlab: bool) -> Option<super::User> {
        use crate::schema::user::dsl::*;
        let _span = metrics::span("find_user_in_db", span);
        user.filter(name.eq(name)).filter(gitlab.eq(_gitlab)).first(&mut self.db).ok()
    }

    pub fn get_users(&mut self) -> Vec<super::User> {
        crate::schema::user::table.load(&mut self.db).expect("Could not load users")
    }

    pub fn save_user(&mut self, span: &Span, user: &super::User) {
        let _span = metrics::span("save_user_to_db", span);
        diesel::update(user).set(user.clone()).execute(&mut self.db).expect("Failed to save user");
    }

    pub fn delete_user(&mut self, user: &super::User) {
        diesel::delete(user).execute(&mut self.db).expect("Failed to delete user");
    }

    // Tokens
    pub fn create_token(&mut self, span: &Span, _owner: i32, _exp: i64) -> super::Token {
        let _span = metrics::span("create_token_in_db", span);
        diesel::insert_into(crate::schema::tokens::table)
            .values(&super::token::NewToken {
                owner_id: _owner,
                exp: _exp
            })
            .get_result(&mut self.db)
            .expect("Failed to save new token to database")
    }

    pub fn get_token(&mut self, span: &Span, _id: i32) -> Option<super::Token> {
        use crate::schema::tokens::dsl::*;
        let _span = metrics::span("get_token_from_db", span);
        tokens.find(_id).first(&mut self.db).ok()
    }

    pub fn delete_token(&mut self, span: &Span, _id: i32) {
        use crate::schema::tokens::dsl::*;
        let _span = metrics::span("delete_token_from_db", span);
        diesel::delete(tokens.find(_id)).execute(&mut self.db).expect("Failed to delete token");
    }

    pub fn delete_all_tokens(&mut self, span: &Span, _owner: i32) {
        use crate::schema::tokens::dsl::*;
        let _span = metrics::span("delete_user_tokens_from_db", span);
        diesel::delete(tokens.filter(owner_id.eq(_owner)))
            .execute(&mut self.db)
            .expect("Failed to delete token");
    }

    pub fn cleanup_tokens(&mut self, span: &Span) {
        use crate::schema::tokens::dsl::*;
        let _span = metrics::span("cleanup_tokens", span);
        let current_time = chrono::Utc::now().timestamp();
        diesel::delete(tokens.filter(exp.le(current_time)))
            .execute(&mut self.db)
            .expect("Failed to cleanup tokens");
    }

    // Nodes
    pub fn get_lock(user: i32) -> super::manager::TracingLock { DB_MANAGER.get_lock(user) }

    pub fn create_node(
        &mut self,
        span: &Span,
        file: bool,
        name: String,
        parent: Option<i32>,
        owner: i32
    ) -> super::Inode {
        DB_MANAGER.create_node(span, &mut self.db, file, name, parent, owner)
    }

    pub fn get_all_nodes(&mut self) -> Vec<super::Inode> {
        crate::schema::inode::table.load(&mut self.db).expect("Could not load nodes")
    }

    pub fn get_node(&mut self, span: &Span, id: i32) -> Option<super::Inode> {
        DB_MANAGER.get_node(span, &mut self.db, id)
    }

    pub fn get_children(&mut self, span: &Span, id: i32) -> Vec<super::Inode> {
        DB_MANAGER.get_children(span, &mut self.db, id)
    }

    pub fn save_node(&mut self, span: &Span, node: &super::Inode) { DB_MANAGER.save_node(span, &mut self.db, node); }

    pub fn delete_node(&mut self, span: &Span, node: &super::Inode) {
        DB_MANAGER.delete_node(span, &mut self.db, node);
    }
}
